! vim:fdm=marker:
!------------------------------------------------------------------------
!> This module contains fast fourier transform wrappers.
!> We use the convention of the FFT without any prefactor (just like FFTW),
!> This implies that the inverse transform of the transform gives the original
!> sequence  times the number of gridpoints, i.e. FFT^+[FFT^-[x]]= N*[x].
!> 
!> Depending on the value of the macro FFTLIB (currently either FFT_FFTW3 or FFT_ACML)
!> different libraries are used in the FFT subroutines. Doxygen gets a little
!> confused here by the preprocessor macros. Check the source code.
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011

#if defined(FFT_FFTW3)
MODULE modulefft
IMPLICIT NONE

CONTAINS

!------------------------------------------------------------------------
!> This is the FFTW3 wrapper for the 1D fft
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011
SUBROUTINE get_FFT1D(instruction,length1,vec,INFO)
!---{{{
IMPLICIT NONE 
INCLUDE 'fftw3.f'

INTEGER*8                       :: plan
INTEGER,INTENT(INOUT)           :: INFO
INTEGER,INTENT(IN)              :: length1,instruction
COMPLEX*16,INTENT(INOUT)        :: vec(length1)


do_what: SELECT CASE(instruction) 
CASE (1) 
  CALL dfftw_plan_dft_1d(plan,length1,vec,vec,FFTW_BACKWARD,FFTW_ESTIMATE)
  CALL dfftw_execute_dft(plan,vec,vec)
CASE (-1)
  CALL dfftw_plan_dft_1d(plan,length1,vec,vec,FFTW_FORWARD,FFTW_ESTIMATE)
  CALL dfftw_execute_dft(plan,vec,vec)
CASE DEFAULT
  WRITE(*,*)"WRONG DIRECTION IN get_FFT1D (FFTW3), instruction",instruction
  STOP
END SELECT do_what

CALL dfftw_destroy_plan(plan)
INFO = 0 !< there is no INFO flag in dfftw_execute_dft
!---}}}
END SUBROUTINE get_FFT1D

!------------------------------------------------------------------------
!> This is the FFTW3 wrapper for the 2D fft
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011
SUBROUTINE get_FFT2D(instruction,length1,length2,vec,INFO)
!---{{{
IMPLICIT NONE 
INCLUDE 'fftw3.f'

INTEGER*8                       :: plan
INTEGER,INTENT(INOUT)           :: INFO
INTEGER,INTENT(IN)              :: length1,length2,instruction
COMPLEX*16,INTENT(INOUT)        :: vec(length1*length2)
COMPLEX*16                      :: vec_FFTW(length1,length2)

vec_FFTW = RESHAPE(vec,(/length1,length2/))

do_what: SELECT CASE(instruction) 
CASE (1) 
  CALL dfftw_plan_dft_2d(plan,length1,length2,vec_FFTW,vec_FFTW,FFTW_BACKWARD,FFTW_ESTIMATE)
  CALL dfftw_execute_dft(plan,vec_FFTW,vec_FFTW)
CASE (-1)
  CALL dfftw_plan_dft_2d(plan,length1,length2,vec_FFTW,vec_FFTW,FFTW_FORWARD,FFTW_ESTIMATE)
  CALL dfftw_execute_dft(plan,vec_FFTW,vec_FFTW)
CASE DEFAULT
  WRITE(*,*)"WRONG DIRECTION IN get_FFT1D (FFTW3), instruction",instruction
  STOP
END SELECT do_what

CALL dfftw_destroy_plan(plan)
INFO = 0 !< there is no INFO flag in dfftw_execute_dft

vec = RESHAPE(vec_FFTW,(/length1*length2/))

!---}}}
END SUBROUTINE get_FFT2D

!------------------------------------------------------------------------
!> This is the FFTW3 wrapper for the 3D fft
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011
SUBROUTINE get_FFT3D(instruction,length1,length2,length3,vec,INFO)
!---{{{
IMPLICIT NONE 
INCLUDE 'fftw3.f'

INTEGER*8                       :: plan
INTEGER,INTENT(INOUT)           :: INFO
INTEGER,INTENT(IN)              :: length1,length2,length3,instruction
COMPLEX*16,INTENT(INOUT)        :: vec(length1*length2*length3)
COMPLEX*16                      :: vec_FFTW(length1,length2,length3)

vec_FFTW = RESHAPE(vec,(/length1,length2,length3/))

do_what: SELECT CASE(instruction) 
CASE (1) 
  CALL dfftw_plan_dft_3d(plan,length1,length2,length3,vec_FFTW,vec_FFTW,FFTW_BACKWARD,FFTW_ESTIMATE)
  CALL dfftw_execute_dft(plan,vec_FFTW,vec_FFTW)
CASE (-1)
  CALL dfftw_plan_dft_3d(plan,length1,length2,length3,vec_FFTW,vec_FFTW,FFTW_FORWARD,FFTW_ESTIMATE)
  CALL dfftw_execute_dft(plan,vec_FFTW,vec_FFTW)
CASE DEFAULT
  WRITE(*,*)"WRONG DIRECTION IN get_FFT1D (FFTW3), instruction",instruction
  STOP
END SELECT do_what

CALL dfftw_destroy_plan(plan)
INFO = 0 !< there is no INFO flag in dfftw_execute_dft

vec = RESHAPE(vec_FFTW,(/length1*length2*length3/))
!---}}}
END SUBROUTINE get_FFT3D


END MODULE modulefft

#elif defined (FFT_ACML)

MODULE modulefft
IMPLICIT NONE

CONTAINS

!------------------------------------------------------------------------
!> This is the ACML wrapper for the 1D fft
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011
SUBROUTINE get_FFT1D(instruction,length1,vec,INFO)
!---{{{
IMPLICIT NONE 

INTEGER,INTENT(INOUT)           :: INFO
INTEGER,INTENT(IN)              :: length1,instruction
COMPLEX*16,INTENT(INOUT)        :: vec(length1)
COMPLEX*16                      :: DUM(length1)

REAL*8                          :: SCALEFAC
COMPLEX*16                      :: COMM(3*length1+100)
LOGICAL                         :: INPL


SCALEFAC        = 1.D0
INPL            = .TRUE.

do_what: SELECT CASE(instruction) 
CASE (1) 
  CALL ZFFT1DX(2,SCALEFAC,INPL,length1,vec,1,DUM,1,COMM,INFO) !< default initialization + bwd transform
CASE (-1)
  CALL ZFFT1DX(-2,SCALEFAC,INPL,length1,vec,1,DUM,1,COMM,INFO) !< default initialization + fwd transform
CASE DEFAULT
  WRITE(*,*)"ILLEGAL INSTRUCTION IN get_FFT1D (acml), instruction",instruction
  STOP
END SELECT do_what

IF (INFO /= 0) THEN
  WRITE(*,*)"1D ZFFT1DX FAILED, INFO",INFO
  STOP
END IF


!---}}}
END SUBROUTINE get_FFT1D

!------------------------------------------------------------------------
!> This is the ACML wrapper for the 2D fft
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011
SUBROUTINE get_FFT2D(instruction,length1,length2,vec,INFO)
!---{{{
IMPLICIT NONE 

INTEGER,INTENT(INOUT)           :: INFO
INTEGER,INTENT(IN)              :: length1,length2,instruction
COMPLEX*16,INTENT(INOUT)        :: vec(length1*length2)
COMPLEX*16                      :: DUM(length1*length2)

REAL*8                          :: SCALEFAC
COMPLEX*16                      :: COMM(length1*length2+3*(length1+length2)+200)
LOGICAL                         :: INPL,LTRANS
INTEGER                         :: INCX1,INCX2,INCY1,INCY2

SCALEFAC        = 1.D0
INPL            = .TRUE.
LTRANS          = .TRUE.
INCX1           =  1
INCX2           =  length1
INCY1           =  1
INCY2           =  1

do_what: SELECT CASE(instruction) 
CASE (1) 
  CALL ZFFT2DX(2,SCALEFAC,LTRANS,INPL,length1,length2,vec,INCX1,INCX2,DUM,INCY1,INCY2,COMM,INFO) !< default initialization + BWD transform
CASE (-1)
  CALL ZFFT2DX(-2,SCALEFAC,LTRANS,INPL,length1,length2,vec,INCX1,INCX2,DUM,INCY1,INCY2,COMM,INFO) !< default initialization + FWD transform
CASE DEFAULT
  WRITE(*,*)"ILLEGAL INSTRUCTION IN get_FFT2D (acml), instruction",instruction
  STOP
END SELECT do_what

IF (INFO /= 0) THEN
  WRITE(*,*)"2D ZFFT2DX FAILED, INFO",INFO
  STOP
END IF
!---}}}
END SUBROUTINE get_FFT2D

!------------------------------------------------------------------------
!> This is the ACML wrapper for the 3D fft
!>
!> \author Kaspar Sakmann (2011)
!> \since 2011
SUBROUTINE get_FFT3D(instruction,length1,length2,length3,vec,INFO)
!---{{{
IMPLICIT NONE 

INTEGER,INTENT(INOUT)           :: INFO
INTEGER,INTENT(IN)              :: length1,length2,length3,instruction
COMPLEX*16,INTENT(INOUT)        :: vec(length1*length2*length3)
COMPLEX*16                      :: DUM(length1*length2*length3)

REAL*8                          :: SCALEFAC
COMPLEX*16                      :: COMM(length1*length2*length3+4*(length1+length2+length3)+300)
LOGICAL                         :: INPL
INTEGER                         :: LCOMM,INCX1,INCX2,INCX3,INCY1,INCY2,INCY3

LCOMM           = length1*length2*length3+4*(length1+length2+length3)+300
SCALEFAC        = 1.D0
INPL            = .TRUE.
INCX1           =  1
INCX2           =  length1
INCX3           =  length1*length2
INCY1           =  1
INCY2           =  1
INCY3           =  1

do_what: SELECT CASE(instruction) 
CASE (1) 
  CALL ZFFT3DY(2,SCALEFAC,INPL,length1,length2,length3,vec,INCX1,INCX2,INCX3,DUM,INCY1,INCY2,INCY3,&
               COMM,LCOMM,INFO) !< default initialization + BWD transform
CASE (-1)
  CALL ZFFT3DY(-2,SCALEFAC,INPL,length1,length2,length3,vec,INCX1,INCX2,INCX3,DUM,INCY1,INCY2,INCY3,&
               COMM,LCOMM,INFO) !< default initialization + FWD transform
CASE DEFAULT
  WRITE(*,*)"ILLEGAL INSTRUCTION IN get_FFT3D (acml), instruction",instruction
  STOP
END SELECT do_what

IF (INFO /= 0) THEN
  WRITE(*,*)"3D ZFFT3DY FAILED, INFO",INFO
  STOP
END IF
!---}}}
END SUBROUTINE get_FFT3D


END MODULE modulefft

#endif
