! vim:fdm=marker:

!> This module contains seven routines which do the 7 integration steps of the MCTDHB algorithm,
!> as it is described in the paper PRA 77, 033613 (2008).
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!> \todo Sort out which variables are IN,INOUT,OUT 
MODULE moduleintegration
IMPLICIT NONE

CONTAINS
!> This subroutine supplies the initial values of the h_kq,W_kqsl,rho_kq,rho_kqsl at t=0 for the 
!> subsequent integration steps.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepOne(nConf,MOrb,NPar,CVec_zero,CData,readPotential,V_ext,&
                      NDX,NDY,NDZ,Dims,gridx,gridy,gridz,AbsTime,nSuperDiags,lambda0,&
                      Psi_zero,BandMat,rho_jk_zero,rho_ijkl_zero)


    USE   modulecomputationcore
    USE   moduleextensions

    INTEGER,INTENT(IN)       :: NDX,NDY,NDZ,MOrb,nConf,nSuperDiags,NPar,Dims
    LOGICAL,INTENT(IN)       :: ReadPotential
    REAL*8,INTENT(IN)        :: lambda0,gridx(NDX),gridy(NDY),gridz(NDZ),AbsTime
    REAL*8,INTENT(INOUT)     :: V_ext(NDX*NDY*NDZ)
    COMPLEX*16,INTENT(INOUT) :: Psi_Zero(NDX*NDY*NDZ,MOrb),CData(MOrb,MOrb,MOrb,MOrb,2)
    COMPLEX*16,INTENT(INOUT) :: rho_ijkl_zero(MOrb,MOrb,MOrb,MOrb),rho_jk_zero(MOrb,MOrb)


    COMPLEX*16,INTENT(INOUT) :: BandMat(nSuperDiags+1,nConf),CVec_zero(nConf)
    !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    !+++++++++++++++++++++++++++++++++++ STEP 1 +++++++++++++++++++++++++++++++++++++++++
    !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ! Given the orbitals Psi at t=AbsTime and the coefficient vector
    ! CVec(AbsTime) compute rho_ijkl and build the Hamiltonian matrix 
    ! write(*,*)"AbsTime",AbsTime
    ! write(*,*)"++++++++ STEP 1 +++++++++++"
     
    CALL get_rho_jk(nConf,MOrb,NPar,CVec_zero,rho_jk_zero)
    CALL get_rho_ijkl(nConf,MOrb,NPar,CVec_zero,rho_ijkl_zero)
    CData(:,:,1,1,1) =  rho_jk_zero 
    CData(:,:,:,:,2) =  rho_ijkl_zero
  
    ! If a time-dependent external potential is used, it needs to be updated here
    IF (.NOT. readPotential) THEN
      CALL get_V_ext(V_ext,NDX,NDY,NDZ,Dims,gridx,gridy,gridz,AbsTime)
    END IF
  
    ! Build the CI matrix H(AbsTime) 
    CALL  get_FullHamiltonianBandMatrix(NPar,MOrb,NDX,NDY,NDZ,Dims,nConf,&
                                        nSuperDiags,lambda0,Psi_zero,BandMat)
END SUBROUTINE do_stepOne 

!> This subroutine propagates the CI vector from t=0 to t=tau/2 using SIL and
!> the matrix elements at t=0. The Matrix elements at t=tau/2 are evaluated.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepTwo(AbsTime,CVec_zero,DtCVec,BandMat,RData,IData,LData,&
                      CVec_tauhalf,NPar,nConf,MOrb,IntPeriod,SILmaxIntOrder,&
                      IntegratorTolError,Relax,CData,tau, &
                      SILIntPeriod1,nSuperDiags,&
                      rho_ijkl_tauhalf,rho_jk_tauhalf,propDirection)


    USE   modulecomputationcore
     
!    EXTERNAL FuncCVecBandMat    

    INTEGER, INTENT(IN) :: nSuperDiags,SILmaxIntOrder,MOrb,nConf,NPar,propDirection
    COMPLEX*16 :: Krylov(nConf,2:SILmaxIntOrder)
    COMPLEX*16,INTENT(INOUT) :: CVec_tauhalf(nConf),DtCVec(nConf)
    COMPLEX*16,INTENT(INOUT) :: CData(MOrb,MOrb,MOrb,MOrb,2)
    COMPLEX*16,INTENT(INOUT) :: BandMat(nSuperDiags+1,nConf),CVec_zero(nConf)
    COMPLEX*16,INTENT(OUT)   ::  rho_ijkl_tauhalf(MOrb,MOrb,MOrb,MOrb),rho_jk_tauhalf(MOrb,MOrb)


    REAL*8, INTENT(INOUT) :: AbsTime,SILIntPeriod1,IntPeriod
    REAL*8, INTENT(INOUT) :: RData(10),tau
    INTEGER*4, INTENT(INOUT) :: IData(30) 
    REAL*8, INTENT(IN) :: IntegratorTolError
    logical, INTENT(INOUT)   :: LData
    logical, INTENT(IN)   :: Relax
    logical   :: RestartSIL,StdForm
    INTEGER    :: TrueOrder,ErrorCode,Steps
    REAL*8     :: EigenVector(0:SILmaxIntOrder,0:SILmaxIntOrder+2), &
                  Diagonal(0:SILmaxIntOrder),OffDiag(SILmaxIntOrder),&
                  EigenVal(0:SILmaxIntOrder),OffDg2(SILmaxIntOrder+1),&
                  norm

    !------------- INITIALIZE  VARIABLES--------------------
    EigenVector = 0.d0
    Diagonal    = 0.d0
    OffDiag     = 0.d0
    EigenVal    = 0.d0
    OffDg2      = 0.d0
    ErrorCode=0
    TrueOrder=0
    Steps=0
    RestartSIL=.TRUE.
    StdForm  = .TRUE.


    ! get DtCVec for SILStep
    IData(1) = propDirection 
    CALL FuncCVecBandMat(AbsTime,CVec_zero,DtCVec,BandMat,RData,IData,LData)

    ! integrate t = 0 -> tau/2
    CVec_tauhalf = CVec_zero
  
    IntPeriod = tau / 2.d0
    CALL SILStep(CVec_tauhalf,DtCVec,nConf,IntPeriod,SILmaxIntOrder,      &
                 IntegratorTolError,Relax,RestartSIL,StdForm,Steps,       &
                 Krylov,SILIntPeriod1,                                    &
                 TrueOrder,ErrorCode,AbsTime,FuncCVecBandMat,             &
                 EigenVector,EigenVal,Diagonal,OffDg2,OffDiag,            &
                 BandMat,RData,IData,LData)                        
  
  
    if (ErrorCode.ne.0) then
      write(*,*)"++++++++ STEP 2 +++++++++++"
      write(*,*)"SILStep ErrorCode =",ErrorCode
      STOP
    end if
  
    !  the integrated time interval may be shorter than IntPeriod if so, adjust tau 
    IntPeriod     = SILIntPeriod1 
    tau           = 2.d0 * SILIntPeriod1
  
    ! normalize CVec to get rid of numerical artefacts 
    CALL normvxz(CVec_tauhalf,norm,nConf) 
    CVec_tauhalf=CVec_tauhalf/norm
  
  
    !  The vector CVec_tauhalf contains CVec(tau/2) where tau/2 = Stepsize  
    !  Note that Stepsize can be less than IntPeriod!
    !  now get rho_jk(tau/2) and rho_jjjj(tau/2) using CVec(tau/2) 
    !  and update CData 
  
    CALL get_rho_jk(nConf,MOrb,NPar,CVec_tauhalf,rho_jk_tauhalf)
    CALL get_rho_ijkl(nConf,MOrb,NPar,CVec_tauhalf,rho_ijkl_tauhalf)
    CData(:,:,1,1,1) =  rho_jk_tauhalf 
    CData(:,:,:,:,2) =  rho_ijkl_tauhalf
  
  
 
END SUBROUTINE do_stepTwo 

!> This subroutine propagates the Orbitals from t=0 to t=tau/2 using ABM and
!> the matrix elements obtained in step two.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepThree(Psi_tauhalf,Psi_zero,Psi_WORK,AbsTime,DtPsi_WORK,&
                        CData,RData,IData,LData,NDX,NDY,NDZ,MOrb,Relax,tau,&
                       IntegratorTolError,restartABM,SILIntPeriod1,&
                       ABMIntOrder) 

    USE   modulecomputationcore

    EXTERNAL AbsABMError

    COMPLEX*16,INTENT(INOUT) :: Psi_Zero(NDX*NDY*NDZ,MOrb),CData(MOrb,MOrb,MOrb,MOrb,2)
    COMPLEX*16,INTENT(INOUT) :: Psi_tauhalf(NDX*NDY*NDZ,MOrb)


    REAL*8, INTENT(INOUT)    :: RData(10),SILIntPeriod1,tau,AbsTime,IntegratorTolError
    INTEGER*4, INTENT(INOUT) :: IData(30)
    LOGICAL, INTENT(INOUT)   :: LData
    LOGICAL, INTENT(IN)      :: Relax,restartABM    

    INTEGER, INTENT(IN)      :: NDX,NDY,NDZ,ABMIntOrder,MOrb
   
    COMPLEX*16 :: Psi_WORK(NDX*NDY*NDZ*MOrb), AuxPsi(NDX*NDY*NDZ*MOrb,ABMIntOrder+1)
    COMPLEX*16 :: DtPsi_WORK(NDX*NDY*NDZ*MOrb)
    INTEGER :: Steps,totalNoGridPts,RepeatedSteps,ErrorCode,k
    REAL*8  :: InitStep,ABMIntPeriod,norm
  
    !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    !+++++++++++++++++++++++++++++++++++ STEP 3 +++++++++++++++++++++++++++++++++++++++++
    !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ! Propagate Psi(0)-> Psi(tau/2)
    ! using rho_jk(tau/2) and rho_ijkl(tau/2) with the ABM integrator
    ! write(*,*)"++++++++ STEP 3 +++++++++++"
    totalNoGridPts=NDX*NDY*NDZ
    Psi_tauhalf = Psi_zero
    Psi_WORK        = RESHAPE(Psi_tauhalf,(/NDX*NDY*NDZ*MOrb/))

    CALL FuncOrbs(AbsTime,Psi_WORK,DtPsi_WORK,CData,RData,IData,LData) ! get dPsi/dt

    ! do the Relaxation/propagation of the orbitals 
    if (Relax) then
      InitStep      = tau
      ABMIntPeriod  = tau
    else 
      InitStep      = SILIntPeriod1
      ABMIntPeriod  = SILIntPeriod1
    end if 

    CALL ABM(Psi_WORK,DtPsi_WORK,totalNoGridPts*MOrb,ABMIntPeriod,AbsTime,ABMIntOrder,   &
             InitStep,IntegratorTolError,restartABM,Steps,RepeatedSteps, &
             ErrorCode,AuxPsi,FuncOrbs,AbsABMError,CData,           &
             RData,IData,LData)
    if(ErrorCode.ne.0) then
      !---{{{ error output
      write(*,*)" ERROR OUTPUT"
      write(6,*)" totalNoGridPts          =    ",totalNoGridPts
      write(6,*)" ABMIntPeriod          =    ",ABMIntPeriod
      write(6,*)" AbsTime            =    ",AbsTime
      write(6,*)" ABMIntOrder        =    ",ABMIntOrder
      write(6,*)" TolError           =    ", 2.d0*IntegratorTolError
      write(6,*)" restartABM         =    ",restartABM
      write(6,*)" ABM Steps          =    ",Steps
      write(6,*)" ABM RepeatedSteps  =    ",RepeatedSteps
      write(6,*)" ABM ErrorCode      =    ",ErrorCode
      read(*,*)
      !----}}}
    end if
    Psi_tauhalf = RESHAPE(Psi_WORK,(/NDX*NDY*NDZ,MOrb/)) !  Psi_WORK -> Psi_tauhalf
 
    do k=1,MOrb

      CALL normvxz(Psi_tauhalf(:,k),norm,NDX*NDY*NDZ) 
      Psi_tauhalf(:,k) = Psi_tauhalf(:,k) / norm 

    end do


END SUBROUTINE do_stepThree 
!> This subroutine propagates the Orbitals from t=tau/2 to t=0 using ABM and
!> the matrix elements obtained in step one for error estimation.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepFour(rho_jk_zero,rho_ijkl_zero,CData,Rdata,IData,LData,&
                       Psi_tildetauhalf,Psi_zero,restartABM,SILIntPeriod1,&
                       AbsTime,IntegratorTolError,NDX,NDY,NDZ,MOrb,ABMIntOrder)
  
      USE   modulecomputationcore

      COMPLEX*16,INTENT(INOUT) :: Psi_tildetauhalf(NDX*NDY*NDZ,MOrb)
      COMPLEX*16,INTENT(INOUT)    :: Psi_zero(NDX*NDY*NDZ,MOrb)
      COMPLEX*16,INTENT(INOUT)    :: rho_ijkl_zero(MOrb,MOrb,MOrb,MOrb),rho_jk_zero(MOrb,MOrb)
      COMPLEX*16,INTENT(INOUT) :: CData(MOrb,MOrb,MOrb,MOrb,2)

      REAL*8, INTENT(INOUT)    :: RData(10),SILIntPeriod1,IntegratorTolError
      INTEGER*4, INTENT(INOUT) :: IData(30)
      LOGICAL, INTENT(INOUT)   :: LData
      LOGICAL, INTENT(IN)   :: restartABM    

      REAL*8, INTENT(INOUT)       :: AbsTime
      INTEGER, INTENT(IN)      :: NDX,NDY,NDZ,MOrb,ABMIntOrder


      COMPLEX*16 :: Psi_WORK(NDX*NDY*NDZ*MOrb), AuxPsi(NDX*NDY*NDZ*MOrb,ABMIntOrder+1)
      COMPLEX*16 :: DtPsi_WORK(NDX*NDY*NDZ*MOrb)
      INTEGER :: Steps,totalNoGridPts,RepeatedSteps,ErrorCode,k
      REAL*8  :: InitStep,ABMIntPeriod,norm

      EXTERNAL AbsABMError

      totalNoGridPts=NDX*NDY*NDZ
      InitStep=SILIntPeriod1
      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      !+++++++++++++++++++++++++++++++++++ STEP 4 +++++++++++++++++++++++++++++++++++++++++
      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      ! propagate Psi_zero->Psi_tildetauhalf using rho_ij_zero and rho_ijkl_zero 
      ! write(*,*)"++++++++ STEP 4 +++++++++++"
      ! change CData accordingly:
 
      CData(:,:,1,1,1) =  rho_jk_zero
      CData(:,:,:,:,2) =  rho_ijkl_zero
      
      ! the ABM integrator requires the time derivative of Psi_zero at the beginning
      ! FuncOrbs normalizes Psi_zero and returns DtPsi
      Psi_tildetauhalf = Psi_zero
     
 
      ! the ABM integrator takes and returns one-dimensional arrays only 
      ! therefore it is necessary to copy the 2d array into a 1d array
      Psi_WORK        = RESHAPE(Psi_tildetauhalf,(/NDX*NDY*NDZ*MOrb/))
 
      CALL FuncOrbs(AbsTime,Psi_WORK,DtPsi_WORK,CData,RData,IData,LData) ! get dPsi/dt
!      write(*,*) SILIntPeriod1
         CALL ABM(Psi_WORK,DtPsi_WORK,totalNoGridPts*MOrb,SILIntPeriod1,AbsTime,ABMIntOrder,&
                     InitStep,IntegratorTolError,restartABM,Steps,RepeatedSteps,&
                     ErrorCode,AuxPsi,FuncOrbs,AbsABMError,CData,               &
                     RData,IData,LData)
      if(ErrorCode.ne.0) then
        !---{{{ error output
        write(*,*)"++++++++ STEP 4 +++++++++++"
        write(*,*)" ERROR OUTPUT"
        write(6,*)" totalNoGridPts          =    ",totalNoGridPts
        write(6,*)" SILIntPeriod1      =    ",SILIntPeriod1
        write(6,*)" AbsTime            =    ",AbsTime
        write(6,*)" ABMIntOrder        =    ",ABMIntOrder
        write(6,*)" TolError           =    ",2.d0*IntegratorTolError
        write(6,*)" restartABM         =    ",restartABM
        write(6,*)" ABM Steps          =    ",Steps
        write(6,*)" ABM RepeatedSteps  =    ",RepeatedSteps
        write(6,*)" ABM ErrorCode      =    ",ErrorCode
        read(*,*)
        !----}}}
      end if
      Psi_tildetauhalf = RESHAPE(Psi_WORK,(/NDX*NDY*NDZ,MOrb/)) !  Psi_WORK -> Psi_tildetauhalf
 
      ! normalize the orbitals 
 
      do k=1,MOrb
 
        CALL normvxz(Psi_tildetauhalf(:,k),norm,NDX*NDY*NDZ) 
        Psi_tildetauhalf(:,k) = Psi_tildetauhalf(:,k) / norm 
        !---{{{error output
        if (ABS(norm-1.d0)>2.d0*IntegratorTolError) then 
          write(*,*)"++++++++ STEP 4 +++++++++++"
          write(*,*)k,"after ABM in STEP 4 norm of Psi_tildetauhalf(:,k)",norm
        end if 
        !---}}}
 
      end do
 
END SUBROUTINE do_stepFour
!> This subroutine propagates the Orbitals from t=tau/2 to t=tau using ABM and
!> the matrix elements obtained in step two. The matrix elements h_kq, W_kqsl at 
!> t=tau are evaluated.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepFive(CData,RData,IData,LData,rho_jk_tauhalf,rho_ijkl_tauhalf,Psi_tau,Psi_tauhalf,&
                       readPotential,V_ext,NDX,NDY,NDZ,MOrb,nConf,nPar,Dims,gridx,gridy,gridz,AbsTime,SILIntPeriod1,&
                      ABMIntOrder,IntegratorTolError,restartABM,nSuperDiags,BandMat,lambda0)

      USE modulecomputationcore
      USE moduleextensions

      LOGICAL,INTENT(IN)       :: ReadPotential
      LOGICAL,INTENT(IN)       :: restartABM

      COMPLEX*16,INTENT(INOUT) :: Psi_tauhalf(NDX*NDY*NDZ,MOrb)
      COMPLEX*16,INTENT(INOUT) :: Psi_tau(NDX*NDY*NDZ,MOrb)
      COMPLEX*16,INTENT(INOUT) :: rho_ijkl_tauhalf(MOrb,MOrb,MOrb,MOrb),rho_jk_tauhalf(MOrb,MOrb)
      COMPLEX*16,INTENT(INOUT) :: CData(MOrb,MOrb,MOrb,MOrb,2)

      COMPLEX*16,INTENT(INOUT) :: BandMat(nSuperDiags+1,nConf)
      REAL*8, INTENT(INOUT)    :: RData(10),SILIntPeriod1,IntegratorTolError
      INTEGER*4, INTENT(INOUT) :: IData(30)
      LOGICAL, INTENT(INOUT)   :: LData

      REAL*8, INTENT(INOUT)    :: AbsTime
      INTEGER, INTENT(IN)      :: NDX,NDY,NDZ,MOrb,nPar,nConf,nSuperDiags,Dims
      INTEGER, INTENT(IN)      :: ABMIntOrder
      REAL*8, INTENT(INOUT)    :: gridx(NDX),gridy(NDY),gridz(NDZ),V_ext(NDX*NDY*NDZ),lambda0

      COMPLEX*16 :: Psi_WORK(NDX*NDY*NDZ*MOrb), AuxPsi(NDX*NDY*NDZ*MOrb,ABMIntOrder+1)
      COMPLEX*16 :: DtPsi_WORK(NDX*NDY*NDZ*MOrb)
      INTEGER :: Steps,totalNoGridPts,RepeatedSteps,ErrorCode,k
      REAL*8  :: InitStep,ABMIntPeriod,norm



      EXTERNAL AbsABMError

      totalNoGridPts=NDX*NDY*NDZ
      InitStep=SILIntPeriod1
      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      !+++++++++++++++++++++++++++++++++++ STEP 5 +++++++++++++++++++++++++++++++++++++++++
      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      ! Propagate Psi(tau/2) -> Psi(tau) using rho_jk(tau/2),rho_ijkl(tau/2) with the ABM integrator
      ! write(*,*)"++++++++ STEP 5 +++++++++++"
 
      ! first adjust CData accordingly
      CData(:,:,1,1,1) =  rho_jk_tauhalf 
      CData(:,:,:,:,2) =  rho_ijkl_tauhalf
 
      ! the ABM integrator requires the time derivative of Psi_zero at the beginning
      ! FuncOrbs normalizes Psi_zero and returns DtPsi
      Psi_tau = Psi_tauhalf
 
      ! the ABM integrator takes and returns one-dimensional arrays only 
      ! therefore it is necessary to copy the 2d array into a 1d array
      Psi_WORK        = RESHAPE(Psi_tau,(/NDX*NDY*NDZ*MOrb/))
 
      ! If a time-dependent external potential is used, it needs to be updated here
      IF (.NOT. readPotential) THEN
        CALL get_V_ext(V_ext,NDX,NDY,NDZ,Dims,gridx,gridy,gridz,AbsTime+SILIntPeriod1)
      END IF
      CALL FuncOrbs(AbsTime+SILIntPeriod1,Psi_WORK,DtPsi_WORK,CData,RData,IData,LData) ! get dPsi/dt
 
      CALL ABM(Psi_WORK,DtPsi_WORK,totalNoGridPts*MOrb,SILIntPeriod1,AbsTime+SILIntPeriod1,&
               ABMIntOrder,InitStep,IntegratorTolError,restartABM, &
               Steps,RepeatedSteps,ErrorCode,AuxPsi,               &
               FuncOrbs,AbsABMError,CData,RData,IData,LData)
 
      if(ErrorCode.ne.0) then
        !---{{{ error output
        write(*,*)"++++++++ STEP 5 +++++++++++"
        write(*,*)" ERROR OUTPUT"
        write(6,*)" totalNoGridPts          =    ",totalNoGridPts
        write(6,*)" SILIntPeriod1      =    ",SILIntPeriod1
        write(6,*)" AbsTime            =    ",AbsTime
        write(6,*)" ABMIntOrder        =    ",ABMIntOrder
        write(6,*)" TolError           =    ",2.d0*IntegratorTolError
        write(6,*)" restartABM         =    ",restartABM
        write(6,*)" ABM Steps          =    ",Steps
        write(6,*)" ABM RepeatedSteps  =    ",RepeatedSteps
        write(6,*)" ABM ErrorCode      =    ",ErrorCode
        read(*,*)
        !----}}}
      end if
 
      Psi_tau = RESHAPE(Psi_WORK,(/NDX*NDY*NDZ,MOrb/)) !  Psi_WORK -> Psi_tau
 
      ! normalize the orbitals
      do k=1,MOrb
        CALL normvxz(Psi_tau(:,k),norm,NDX*NDY*NDZ) 
        Psi_tau(:,k) = Psi_tau(:,k) / norm 
      !---{{{error output
        if (ABS(norm-1.d0)>2.d0*IntegratorTolError) then 
          write(*,*)"++++++++ STEP 5 +++++++++++"
          write(*,*)k,"after ABM in STEP 5 norm of Psi_tau(:,k)",norm
        end if 
      !---}}}
      end do
 
      ! If a time-dependent external potential is used, it needs to be updated here
      IF (.NOT. readPotential) THEN
        CALL get_V_ext(V_ext,NDX,NDY,NDZ,Dims,gridx,gridy,gridz,AbsTime+2*SILIntPeriod1)
      END IF
      CALL  get_FullHamiltonianBandMatrix(NPar,MOrb,NDX,NDY,NDZ,Dims,nConf,&
                                         nSuperDiags,lambda0,Psi_tau,BandMat)
      
 

END SUBROUTINE do_stepFive

!> This subroutine propagates the CI vector from t=tau/2 to t=tau using SIL and
!> the matrix elements obtained in step five.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepSix(CVec_tau,CVec_tauhalf,SILIntPeriod1,BandMat,CData,RData,&
                      IData,LData,nConf,SILmaxIntOrder,IntegratorTolError,Relax,RestartSIL,&
                      StdForm,SILIntPeriod2,AbsTime,iFail,MOrb,nSuperDiags,tau)


    USE modulecomputationcore


    INTEGER, INTENT(IN) :: nSuperDiags,SILmaxIntOrder,MOrb,nConf
    INTEGER, INTENT(INOUT) :: iFail
    COMPLEX*16 :: Krylov(nConf,2:SILmaxIntOrder),DtCVec(nConf)
    COMPLEX*16,INTENT(INOUT) :: CVec_tauhalf(nConf)
    COMPLEX*16,INTENT(INOUT) :: CVec_tau(nConf)
    COMPLEX*16,INTENT(INOUT) :: CData(MOrb,MOrb,MOrb,MOrb,2)
    COMPLEX*16,INTENT(INOUT) :: BandMat(nSuperDiags+1,nConf)

    REAL*8, INTENT(INOUT) :: AbsTime,SILIntPeriod1,SILIntPeriod2
    REAL*8, INTENT(INOUT) :: RData(10),tau
    INTEGER*4, INTENT(INOUT) :: IData(30) 
    REAL*8, INTENT(IN) :: IntegratorTolError
    logical, INTENT(INOUT)   :: LData
    logical, INTENT(IN)   :: Relax
 
    logical    :: RestartSIL,StdForm
    INTEGER    :: TrueOrder,ErrorCode,Steps
    REAL*8     :: EigenVector(0:SILmaxIntOrder,0:SILmaxIntOrder+2), &
                  Diagonal(0:SILmaxIntOrder),OffDiag(SILmaxIntOrder),&
                  EigenVal(0:SILmaxIntOrder),OffDg2(SILmaxIntOrder+1),&
                  norm,IntPeriod

    !------------- INITIALIZE  VARIABLES--------------------
    EigenVector = 0.d0
    Diagonal    = 0.d0
    OffDiag     = 0.d0
    EigenVal    = 0.d0
    OffDg2      = 0.d0
    ErrorCode=0
    TrueOrder=0
    Steps=0
    RestartSIL=.TRUE.
    StdForm  = .TRUE.


      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      !+++++++++++++++++++++++++++++++++++ STEP 6 +++++++++++++++++++++++++++++++++++++++++
      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      ! Propagate CVec(tau/2)->CVec(tau) using H(tau)
      ! write(*,*)"++++++++ STEP 6 +++++++++++"
      CVec_tau = CVec_tauhalf
      CALL FuncCVecBandMat(AbsTime+2*SILIntPeriod1,CVec_tau,DtCVec,BandMat,RData,IData,LData)
 
      ! integrate t = tau/2 -> tau
      IntPeriod = SILIntPeriod1
      CALL SILStep(CVec_tau,DtCVec,nConf,IntPeriod,SILmaxIntOrder,IntegratorTolError, &
                   Relax,RestartSIL,StdForm,Steps,Krylov,SILIntPeriod2,       &
                   TrueOrder,ErrorCode,AbsTime,FuncCVecBandMat,               &
                   EigenVector,EigenVal,Diagonal,OffDg2,OffDiag,              &
                   BandMat,RData,IData,LData)                        
 
      if (ErrorCode.ne.0) then
        write(*,*)"++++++++ STEP 6 +++++++++++"
        write(*,*)"SILStep ErrorCode =",ErrorCode
        STOP
      end if
 
      ! Note that SILIntPeriod2 can be less than IntPeriod!
      ! compare the integration periods of the SIL integrations
      if(Abs(SILIntPeriod1-SILIntPeriod2)>1.d-10) then
 
        write(*,*)"The SIL integration Periods of Step 2 and 6 are different"
        write(*,*)"SILIntPeriod1",SILIntPeriod1
        write(*,*)"SILIntPeriod2",SILIntPeriod2
        write(*,*)"repeat the step with a smaller step size"
        iFail = 1
        read(*,*)
 
      end if
 
      ! make sure CVec_tau is normalized
      CALL normvxz(CVec_tau,norm,nConf)
      CVec_tau = CVec_tau / norm
 

END SUBROUTINE do_stepSix

!> This subroutine propagates the CI vector from t=tau/2 to t=0 using SIL and
!> the matrix elements obtained in step five for error estimation.
!> The error is given by the differences of the CI vector of this step to the
!> CI vector of step one and the difference of the orbitals obtained 
!> in step three and step four.
!>
!> \authors Kaspar Sakmann, Axel Lode (2011)
!> \since 2007
!
SUBROUTINE do_stepSeven(CVec_tildezero,CVec_tauhalf,AbsTime,SILIntPeriod1,CVec_tau,BandMat,&
                        CData,RData,IData,LData,nConf,MOrb,propDirection,nSuperDiags,SILmaxIntOrder,&
                        IntegratorTolError,Stepsize,iFail,Relax)

    use modulecomputationcore

    INTEGER, INTENT(IN) :: nSuperDiags,SILmaxIntOrder,MOrb,nConf
    COMPLEX*16 :: Krylov(nConf,2:SILmaxIntOrder),DtCVec(nConf)
    COMPLEX*16,INTENT(INOUT) :: CVec_tauhalf(nConf),CVec_tau(nConf)
    COMPLEX*16,INTENT(INOUT) :: CVec_tildezero(nConf)
    COMPLEX*16,INTENT(INOUT) :: CData(MOrb,MOrb,MOrb,MOrb,2)
    COMPLEX*16,INTENT(INOUT) :: BandMat(nSuperDiags+1,nConf)
    REAL*8, INTENT(INOUT) :: AbsTime,SILIntPeriod1
    REAL*8, INTENT(INOUT) :: RData(10),Stepsize
    INTEGER*4, INTENT(INOUT) :: IData(30) 
    REAL*8, INTENT(IN) :: IntegratorTolError
    logical, INTENT(INOUT)   :: LData,Relax
    INTEGER, INTENT(INOUT)   :: iFail,propDirection

    logical    :: RestartSIL,StdForm
    INTEGER    :: TrueOrder,ErrorCode,Steps
    REAL*8     :: EigenVector(0:SILmaxIntOrder,0:SILmaxIntOrder+2), &
                  Diagonal(0:SILmaxIntOrder),OffDiag(SILmaxIntOrder),&
                  EigenVal(0:SILmaxIntOrder),OffDg2(SILmaxIntOrder+1),&
                  norm,IntPeriod
    !------------- INITIALIZE  VARIABLES--------------------
    EigenVector = 0.d0
    Diagonal    = 0.d0
    OffDiag     = 0.d0
    EigenVal    = 0.d0
    OffDg2      = 0.d0
    ErrorCode=0
    TrueOrder=0
    Steps=0
    RestartSIL=.TRUE.
    StdForm  = .TRUE.




      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      !+++++++++++++++++++++++++++++++++++ STEP 7 +++++++++++++++++++++++++++++++++++++++++
      !++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      ! Propagate backwards CVec(tau/2)->CVe_tilde(0) using H(tau) for error
      ! estimation
      ! write(*,*)"++++++++ STEP 7 +++++++++++"
      ! get DtCVec for SILStep
      CVec_tildezero = CVec_tauhalf
 
      IData(1) = (-1)*propDirection !backward propagation
      CALL FuncCVecBandMat(AbsTime+2.d0*SILIntPeriod1,CVec_tau,DtCVec,BandMat,RData,IData,LData)
 
      ! integrate t = tau/2 -> 0
      IntPeriod = SILIntPeriod1
      CALL SILStep(CVec_tildezero,DtCVec,nConf,IntPeriod,                   & 
                   SILmaxIntOrder,IntegratorTolError,                       &
                   Relax,RestartSIL,StdForm,Steps,Krylov,Stepsize,          &
                   TrueOrder,ErrorCode,AbsTime,FuncCVecBandMat,             &
                   EigenVector,EigenVal,Diagonal,OffDg2,OffDiag,            &
                   BandMat,RData,IData,LData)                        
      !---{{{ error output
      if (ErrorCode.ne.0) then
        write(*,*)"++++++++ STEP 7 +++++++++++"
        write(*,*)"SILStep ErrorCode =",ErrorCode
        STOP
      end if
 
 
      if (Abs(Stepsize-SILIntPeriod1)>1.d-10) then
        write(*,*)"++++++++ STEP 7 +++++++++++"
        write(*,*)"backward integration had a different length than the forward integration in STEP 2"
        write(*,*)"repeat the step with a smaller step size"
        write(*,*)"SILIntPeriod1",SILIntPeriod1
        write(*,*)"Stepsize",Stepsize
        iFail = 2
      end if   
      !---}}}
 
 
END SUBROUTINE do_stepSeven

END MODULE moduleintegration
